//
//  CameraKitFlutterView.swift
//  camerakit
//
//  Created by faranegar on 9/6/20.
//

import Foundation
import AVFoundation
import MLKitBarcodeScanning
import MLKitCommon
import MLKitVision


@available(iOS 10.0, *)
class CameraKitFlutterView : NSObject, FlutterPlatformView, AVCaptureVideoDataOutputSampleBufferDelegate{
    let channel: FlutterMethodChannel
    let frame: CGRect
    
    var takePicture:Bool = false
    var hasBarcodeReader:Bool!
    var previewFlashMode:String!
    
    var previewView : UIView!
    var videoDataOutput: AVCaptureVideoDataOutput!
    var videoDataOutputQueue: DispatchQueue!
    var previewLayer:AVCaptureVideoPreviewLayer!
    var captureDevice : AVCaptureDevice!
    let session = AVCaptureSession()
    var barcodeScanner:BarcodeScanner!
    var flutterResultTakePicture:FlutterResult!
    
    init(registrar: FlutterPluginRegistrar, viewId: Int64, frame: CGRect) {
         self.channel = FlutterMethodChannel(name: "plugins/camera_kit_" + String(viewId), binaryMessenger: registrar.messenger())
        self.frame = frame
         

    
     }
    
    
    
    public func setMethodHandler() {
        self.channel.setMethodCallHandler({ (FlutterMethodCall,  FlutterResult) in
                let args = FlutterMethodCall.arguments
                let myArgs = args as? [String: Any]
                if FlutterMethodCall.method == "requestPermission" {
                    FlutterResult(true)
                } else if FlutterMethodCall.method == "initCamera" {
            
                    self.initCamera(hasBarcodeReader: (myArgs?["hasBarcodeReader"] as! Bool),
                                    flashMode: (myArgs?["flashMode"] ) as! String)
                } else if FlutterMethodCall.method == "resumeCamera" {
                    self.beginSession()
            }
                else if FlutterMethodCall.method == "pauseCamera" {
                    self.stopCamera()
                }
            else if FlutterMethodCall.method == "changeFlashMode" {
//                    self.initCamera()
                    self.changeFlashMode(flashMode: (myArgs?["flashMode"] ) as! String)
            }
             else if FlutterMethodCall.method == "takePicture" {
                    self.flutterResultTakePicture = FlutterResult
                    self.takePicture(flashMode: "O")
                        }
            })
    }
    
    func changeFlashMode(flashMode: String) {
        if flashMode == "A" {
            flashAuto()
        }
        else if flashMode == "O" {
            flashOn()
        } else if flashMode == "F"{
            flashOff()
        }
    }
    
    func view() -> UIView {
        if previewView == nil {
        self.previewView = UIView(frame: frame)
         previewView.contentMode = UIView.ContentMode.scaleAspectFit
         
//            previewView.backgroundColor = UIColor(red: 125, green: 200, blue: 70, alpha: 1)
        }
        return previewView
            
              //Add a view on top of the cameras' view
         

              //self.setupAVCapture()
    }
    
    func initCamera(hasBarcodeReader: Bool, flashMode: String) {
        self.hasBarcodeReader = hasBarcodeReader
        self.previewFlashMode = flashMode
        if hasBarcodeReader == true{
            let format = BarcodeFormat.all
                let barcodeOptions = BarcodeScannerOptions(formats: format)

                // Create a barcode scanner.
               barcodeScanner = BarcodeScanner.barcodeScanner(options: barcodeOptions)
        }
        DispatchQueue.main.asyncAfter(deadline: .now() + 0.5) {
            self.setupAVCapture()
        }
        
 
    }
    
    @available(iOS 10.0, *)
    func setupAVCapture(){
          session.sessionPreset = AVCaptureSession.Preset.vga640x480
          guard let device = AVCaptureDevice
          .default(AVCaptureDevice.DeviceType.builtInWideAngleCamera,
                   for: .video,
                   position: AVCaptureDevice.Position.back) else {
                              return
          }
          captureDevice = device
          changeFlashMode(flashMode: previewFlashMode)
          beginSession()
      }
    
    private func flashAuto()
    {
        do{
            if (captureDevice.hasTorch)
            {
                try captureDevice.lockForConfiguration()
                captureDevice.torchMode = .auto
                captureDevice.flashMode = .auto
                captureDevice.unlockForConfiguration()
            }
        }catch{
            //DISABEL FLASH BUTTON HERE IF ERROR
            print("Device tourch Flash Error ");
        }
    }
    
    private func flashOn()
    {
        do{
            if (captureDevice.hasTorch)
            {
                try captureDevice.lockForConfiguration()
                captureDevice.torchMode = .on
                captureDevice.flashMode = .on
                captureDevice.unlockForConfiguration()
            }
        }catch{
            //DISABEL FLASH BUTTON HERE IF ERROR
            print("Device tourch Flash Error ");
        }
    }
    
    private func flashOff()
       {
           do{
               if (captureDevice.hasTorch){
                   try captureDevice.lockForConfiguration()
                   captureDevice.torchMode = .off
                   captureDevice.flashMode = .off
                   captureDevice.unlockForConfiguration()
               }
           }catch{
               //DISABEL FLASH BUTTON HERE IF ERROR
               print("Device tourch Flash Error ");
           }
       }
    
    func beginSession(){
        var deviceInput: AVCaptureDeviceInput!

        
        do {
            deviceInput = try AVCaptureDeviceInput(device: captureDevice)
            guard deviceInput != nil else {
                print("error: cant get deviceInput")
                return
            }

            if self.session.canAddInput(deviceInput){
                self.session.addInput(deviceInput)
            }

            videoDataOutput = AVCaptureVideoDataOutput()
            videoDataOutput.alwaysDiscardsLateVideoFrames=true
            videoDataOutputQueue = DispatchQueue(label: "VideoDataOutputQueue")
            videoDataOutput.setSampleBufferDelegate(self, queue:self.videoDataOutputQueue)

            if session.canAddOutput(self.videoDataOutput){
                session.addOutput(self.videoDataOutput)
            }

            videoDataOutput.connection(with: .video)?.isEnabled = true

            previewLayer = AVCaptureVideoPreviewLayer(session: self.session)
            previewLayer.videoGravity = AVLayerVideoGravity.resizeAspect

            let rootLayer :CALayer = self.previewView.layer
            rootLayer.masksToBounds=true
            previewLayer.frame = rootLayer.bounds
            rootLayer.addSublayer(self.previewLayer)
            session.startRunning()
        } catch let error as NSError {
            deviceInput = nil
            print("error: \(error.localizedDescription)")
        }
    }
    
    func stopCamera(){
        session.stopRunning()
    }

    
     private func currentUIOrientation() -> UIDeviceOrientation {
        let deviceOrientation = { () -> UIDeviceOrientation in
          switch UIApplication.shared.statusBarOrientation {
          case .landscapeLeft:
            return .landscapeRight
          case .landscapeRight:
            return .landscapeLeft
          case .portraitUpsideDown:
            return .portraitUpsideDown
          case .portrait, .unknown:
            return .portrait
          @unknown default:
            fatalError()
          }
        }
        guard Thread.isMainThread else {
          var currentOrientation: UIDeviceOrientation = .portrait
          DispatchQueue.main.sync {
            currentOrientation = deviceOrientation()
          }
          return currentOrientation
        }
        return deviceOrientation()
      }
    
    
    public func imageOrientation(
      fromDevicePosition devicePosition: AVCaptureDevice.Position = .back
    ) -> UIImage.Orientation {
      var deviceOrientation = UIDevice.current.orientation
      if deviceOrientation == .faceDown || deviceOrientation == .faceUp
        || deviceOrientation
          == .unknown
      {
        deviceOrientation = currentUIOrientation()
      }
      switch deviceOrientation {
      case .portrait:
        return devicePosition == .front ? .leftMirrored : .right
      case .landscapeLeft:
        return devicePosition == .front ? .downMirrored : .up
      case .portraitUpsideDown:
        return devicePosition == .front ? .rightMirrored : .left
      case .landscapeRight:
        return devicePosition == .front ? .upMirrored : .down
      case .faceDown, .faceUp, .unknown:
        return .up
      @unknown default:
        fatalError()
      }
    }
    func saveImage(image: UIImage) -> Bool {
        guard let data = image.jpegData(compressionQuality: 1) ?? image.pngData() else {
            return false
        }
        guard let directory = try? FileManager.default.url(for: .documentDirectory, in: .userDomainMask, appropriateFor: nil, create: false) as NSURL else {
            return false
        }
        do {
            try data.write(to: directory.appendingPathComponent("pic.jpg")!)
            flutterResultTakePicture(directory.path!  + "/pic.jpg")
            //print(directory)
            return true
        } catch {
            print(error.localizedDescription)
            return false
        }
    }
    func takePicture(flashMode:Character) {
        self.takePicture = true
    }

    func captureOutput(_ output: AVCaptureOutput, didOutput sampleBuffer: CMSampleBuffer, from connection: AVCaptureConnection) {
        
        if(self.takePicture) {
            guard let cvBuffer = CMSampleBufferGetImageBuffer(sampleBuffer) else {
                    return
                }
                
                //get a CIImage out of the CVImageBuffer
                let ciImage = CIImage(cvImageBuffer: cvBuffer)
                
                //get UIImage out of CIImage
                let uiImage = UIImage(ciImage: ciImage)
                saveImage(image: uiImage)
            
            
            self.takePicture = false
            return
        }
        
        // do stuff here
        if barcodeScanner != nil {
            let visionImage = VisionImage(buffer: sampleBuffer)
            let orientation = imageOrientation(
              fromDevicePosition: .back
            )
            visionImage.orientation = orientation
            var barcodes: [Barcode]
            do {
                barcodes = try self.barcodeScanner.results(in: visionImage)
            } catch let error {
              print("Failed to scan barcodes with error: \(error.localizedDescription).")
              return
            }
            
            guard !barcodes.isEmpty else {
               //print("Barcode scanner returrned no results.")
               return
             }
            
            for barcode in barcodes {
                barcodeRead(barcode: barcode.rawValue!)
            }
        }
        
    }
    
    func barcodeRead(barcode: String) {
        channel.invokeMethod("onBarcodeRead", arguments: barcode)
    }
    
}
